#![no_main]
#![no_std]

extern crate cortex_m;
extern crate cortex_m_rt as rt;
extern crate panic_semihosting;

extern crate stm32f4xx_hal as hal;

mod packet;

#[rtic::app(device = stm32f4xx_hal::pac, peripherals = true)]
mod app {
    use hal::{
        block,
        dma::{
            config::DmaConfig,
            traits::{Stream, StreamISR},
            PeripheralToMemory, Stream2, StreamsTuple, Transfer,
        },
        gpio,
        pac::{DMA2, SPI1, TIM5, USART1},
        prelude::*,
        serial::{self, Config, Rx, Serial, Tx},
        spi::{Mode, Phase, Polarity, Spi},
        timer::Delay,
    };
    use rf4463::{config::RADIO_CONFIG_500_2, Rf4463};
    use ringbuffer::{ConstGenericRingBuffer, RingBuffer, RingBufferRead, RingBufferWrite};
    use rtic::Mutex;

    use crate::packet::Packet;

    // in # packets
    const BUFFER_LEN: usize = 50;
    // # bytes
    // if we make this too big, the interrupt may take too long to process them all
    // and we could drop packets!
    // ideally we would use the rf4463 interrupt (higher priority)
    // to keep the transceiver fed
    const DMA_BUFFER_LEN: usize = 128;

    const MODE: Mode = Mode {
        polarity: Polarity::IdleLow,
        phase: Phase::CaptureOnFirstTransition,
    };

    type Radio = Rf4463<
        Spi<SPI1>,
        gpio::Pin<'B', 1, gpio::Output>,
        gpio::Pin<'B', 0, gpio::Output>,
        Delay<TIM5, 1000000>,
    >;

    type RxTransfer = Transfer<
        Stream2<DMA2>,
        4,
        Rx<USART1>,
        PeripheralToMemory,
        &'static mut [u8; DMA_BUFFER_LEN],
    >;

    #[derive(Debug)]
    enum SlaveCmd {
        BufferStatus,
        SendPacket,
        GetTemp,
    }

    impl SlaveCmd {
        pub fn from_u8(i: u8) -> Option<Self> {
            match i {
                0x00 => Some(Self::BufferStatus),
                0x01 => Some(Self::SendPacket),
                0x02 => Some(Self::GetTemp),
                _ => None,
            }
        }
    }

    #[allow(clippy::large_enum_variant)]
    pub enum SlaveState {
        Idle,
        RecvPacket(Packet),
    }

    #[shared]
    struct Shared {
        radio_temp: f32,
        tx_buf: ConstGenericRingBuffer<Packet, BUFFER_LEN>,

        // Can use lock free for any resources that are only shared between tasks of the same priority
        #[lock_free]
        state: SlaveState,

        #[lock_free]
        tx: Tx<USART1>,

        #[lock_free]
        rx_transfer: RxTransfer,

        #[lock_free]
        rx_buffer: Option<&'static mut [u8; DMA_BUFFER_LEN]>,
    }

    #[local]
    struct Local {
        radio: Radio,
    }

    #[init]
    fn init(ctx: init::Context) -> (Shared, Local, init::Monotonics) {
        let rcc = ctx.device.RCC.constrain();

        let gpioa = ctx.device.GPIOA.split();
        let gpiob = ctx.device.GPIOB.split();

        // setup clocks
        let clocks = rcc
            .cfgr
            .use_hse(25.MHz())
            .sysclk(100.MHz())
            .pclk1(48.MHz())
            .pclk2(48.MHz())
            .freeze();

        // setup 4463 spi
        let mosi = gpioa.pa7.into_alternate();
        let miso = gpioa.pa6.into_alternate();
        let sclk = gpioa.pa5.into_alternate();

        let sdn = gpiob.pb1.into_push_pull_output();

        let cs = gpiob.pb0.into_push_pull_output();

        let spi = Spi::new(ctx.device.SPI1, (sclk, miso, mosi), MODE, 10.MHz(), &clocks);

        let delay = ctx.device.TIM5.delay_us(&clocks);
        let radio = Rf4463::new(spi, sdn, cs, delay, &mut RADIO_CONFIG_500_2.clone()).unwrap();

        // setup Pi UART
        let tx = gpioa.pa9;
        let rx = gpioa.pa10;

        let usart = Serial::new(
            ctx.device.USART1,
            (tx, rx),
            Config::default()
                .baudrate(921_600.bps())
                .dma(serial::config::DmaConfig::Rx),
            &clocks,
        )
        .unwrap();

        let (tx, mut rx) = usart.split();

        rx.listen_idle();

        let dma2 = StreamsTuple::new(ctx.device.DMA2);

        let rx_buffer1 =
            cortex_m::singleton!(: [u8; DMA_BUFFER_LEN] = [0; DMA_BUFFER_LEN]).unwrap();
        let rx_buffer2 =
            cortex_m::singleton!(: [u8; DMA_BUFFER_LEN] = [0; DMA_BUFFER_LEN]).unwrap();

        // todo figure out double buffering
        let mut rx_transfer = Transfer::init_peripheral_to_memory(
            dma2.2,
            rx,
            rx_buffer1,
            None,
            DmaConfig::default()
                .memory_increment(true)
                .fifo_enable(true)
                .fifo_error_interrupt(true)
                .transfer_complete_interrupt(true),
        );

        rx_transfer.start(|_rx| {});

        (
            Shared {
                radio_temp: 0.0,
                tx_buf: ConstGenericRingBuffer::new(),
                tx,
                state: SlaveState::Idle,

                rx_transfer,
                rx_buffer: Some(rx_buffer2),
            },
            Local { radio },
            init::Monotonics(),
        )
    }

    #[idle(local=[radio], shared=[tx_buf, radio_temp])]
    fn idle(mut ctx: idle::Context) -> ! {
        let mut i = 0;
        loop {
            if let Some(mut pkt) = ctx.shared.tx_buf.lock(|buf| buf.dequeue()) {
                ctx.local.radio.tx(&mut pkt.data).unwrap();
            }

            i += 1;

            // get temp every 500 packets, or whenever when idle
            if i >= 500 {
                i = 0;

                if let Ok(new_temp) = ctx.local.radio.get_temp() {
                    ctx.shared.radio_temp.lock(|cur_temp| *cur_temp = new_temp);
                }
            }
        }
    }

    #[task(binds = USART1, priority=1, shared = [state, tx_buf, tx,radio_temp, rx_transfer, rx_buffer])]
    fn usart1(mut ctx: usart1::Context) {
        let transfer = &mut ctx.shared.rx_transfer;

        if transfer.is_idle() {
            let byte_count =
                DMA_BUFFER_LEN - usize::from(Stream2::<DMA2>::get_number_of_transfers());
            let new_buf = ctx.shared.rx_buffer.take().unwrap();
            let (buffer, _) = transfer.next_transfer(new_buf).unwrap();
            let bytes = &buffer[..byte_count];

            handle_bytes(
                ctx.shared.state,
                ctx.shared.tx_buf,
                ctx.shared.tx,
                ctx.shared.radio_temp,
                bytes,
            );

            *ctx.shared.rx_buffer = Some(buffer);
        }
    }

    // Must be same priority as USART1
    #[task(binds = DMA2_STREAM2, priority=1, shared = [state, tx_buf, tx, radio_temp, rx_transfer, rx_buffer])]
    fn dma2_stream2(mut ctx: dma2_stream2::Context) {
        let transfer = &mut ctx.shared.rx_transfer;

        if Stream2::<DMA2>::get_fifo_error_flag() {
            transfer.clear_fifo_error_interrupt();
        }
        if Stream2::<DMA2>::get_transfer_complete_flag() {
            transfer.clear_transfer_complete_interrupt();

            let new_buf = ctx.shared.rx_buffer.take().unwrap();
            let (buffer, _) = transfer.next_transfer(new_buf).unwrap();

            handle_bytes(
                ctx.shared.state,
                ctx.shared.tx_buf,
                ctx.shared.tx,
                ctx.shared.radio_temp,
                buffer,
            );

            *ctx.shared.rx_buffer = Some(buffer);
        }
    }

    fn handle_bytes<
        B: Mutex<T = ConstGenericRingBuffer<Packet, BUFFER_LEN>>,
        RT: Mutex<T = f32>,
    >(
        state: &mut SlaveState,
        mut buf: B,
        tx: &mut Tx<USART1>,
        mut radio_temp: RT,
        bytes: &[u8],
    ) {
        for &x in bytes {
            match state {
                SlaveState::Idle => {
                    if let Some(cmd) = SlaveCmd::from_u8(x) {
                        match cmd {
                            SlaveCmd::BufferStatus => {
                                if buf.lock(|b| b.is_full()) {
                                    block!(tx.write(0x01)).unwrap(); // not enough space for a new packet
                                } else {
                                    block!(tx.write(0x02)).unwrap(); // safe to send a packet
                                }
                            }

                            SlaveCmd::SendPacket => {
                                if buf.lock(|b| b.len() >= b.capacity() - 1) {
                                    block!(tx.write(0x01)).unwrap(); // not enough space for a new packet
                                } else {
                                    block!(tx.write(0x02)).unwrap(); // safe to send a packet
                                }

                                *state = SlaveState::RecvPacket(Packet::default());
                            }

                            SlaveCmd::GetTemp => {
                                let temp = radio_temp.lock(|x| *x);
                                tx.bwrite_all(&temp.to_le_bytes()).unwrap();
                            }
                        }
                    }
                }

                SlaveState::RecvPacket(pkt) => {
                    pkt.push(x);
                    if pkt.is_complete() {
                        buf.lock(|b| b.enqueue(pkt.clone()));

                        *state = SlaveState::Idle;
                    }
                }
            };
        }
    }
}
